/*
 *  Copyright (c) 2005, FlexiGuided GmbH, Berlin, Germany
 *  Author: Jan Behrens <jan.behrens@flexiguided.de>
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the FlexiGuided GmbH nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 *  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
 

/*
 *  File name:    errhdl.h
 *  Version:      0.2
 *  Last changed: 2005-12-13
 *
 *  Dependencies: [none]
 *
 *  Description:
 *  C header file for error handling library, including macro definitions.
 *
 *  WARNING: Using compilers for comiling either the error handling library
 *  or programs using the library, which optimize the generated code, is
 *  dangerous!
 *  USE OPTIMIZED COMPILATION ONLY IF YOU KNOW WHAT YOU ARE DOING!
 *  For disabling optimization for the GNU C compiler search for the -O0
 *  switch in the manpage of the GNU C compiler.
 *
 */


#ifndef ERRHDL_H
#define ERRHDL_H

#include <setjmp.h>
#include <string.h>
#include <stdbool.h>

/*
 *  'ERRHDL_MSGLEN' is the maximum length of an error message.
 *  Longer messages will be truncated.
 */
#define ERRHDL_MSGLEN 400

/*
 *  'errhdl_panic_fptr' is the type of a function
 *  being called in case of an unhandled error.
 */
typedef void (*errhdl_panic_fptr)(char *class, char *message);

/*
 *  'errhdl_state_t' is the type of a structure
 *  storing error handling context variables.
 */
typedef struct errhdl_state_struct {
	errhdl_panic_fptr panic;
	jmp_buf *envptr;
	char *class;
	char message[ERRHDL_MSGLEN+1];
} errhdl_state_t;

/*
 *  'errhdl_state_get_fptr' is the type of a function
 *  returning a pointer to the current error handling context.
 */
typedef errhdl_state_t *(*errhdl_state_get_fptr)();

/*
 *  'errhdl_state_global' is the global error handling context,
 *  'errhdl_state_get' points to a function returning the current error
 *  handling context, and 'errhdl_state' is a macro for invoking either a
 *  function to get the current error handling context or using the global
 *  context, if no function to do so is available.
 */
extern errhdl_state_t errhdl_state_global;
extern errhdl_state_get_fptr errhdl_state_get;
#define errhdl_state ( \
	errhdl_state_get ? errhdl_state_get() : &errhdl_state_global \
	)

/*
 *  The 'ErrHdl' macro reserves memory for a pointer to an error class
 *  and for a buffer for an error message on the stack.
 */
#define ErrHdl \
	char *errhdl_class = 0; \
	char errhdl_message[ERRHDL_MSGLEN+1] = ""

/*
 *  'Try' indicates that the following command or block is to be executed in
 *  a protected way. If any error occurs, execution continues on the
 *  'OnError' statement, which MUST follow a 'Try' statement.
 */
#define Try \
	{ \
	jmp_buf errhdl_env, *errhdl_oldenvptr; \
	errhdl_oldenvptr = errhdl_state->envptr; \
	errhdl_state->envptr = &errhdl_env; \
	errhdl_state->class = 0; \
	if (setjmp(errhdl_env) == 0)

/*
 *  'OnError' MUST follow a 'Try' statement, and indicates that the
 *  following block or command (if existent) should be executed
 *  in case of error. An 'else' may follow.
 */
#define OnError \
	errhdl_state->envptr = errhdl_oldenvptr; \
	} \
	if ((errhdl_class = errhdl_state->class)) { \
		strncpy(errhdl_message, errhdl_state->message, ERRHDL_MSGLEN); \
		errhdl_message[ERRHDL_MSGLEN] = 0; \
	} else errhdl_message[0] = 0; \
	if (errhdl_class)

/*
 *  'errhdl_error' and 'errhdl_errorofclass' are to be invoked by macros
 *  and are not supposed to be invoked directly.
 */ 
extern void errhdl_error(char *class, char *msgfmt, ...);
extern bool errhdl_errorofclass(char *subclass, char *superclass);

/*
 *  'Error' raises an error condition, and jumps to the 'OnError' statement
 *  or aborts the program execution, if it was not called inside a 'Try'
 *  block.
 *  'Escalate' raises the last error (respective to the 'ErrHdl'
 *  declaration) again.
 *  'ErrorOfClass' checks if the last error occured is of the given class.
 *  'ErrorMessage' is a pointer to the last error message.
 */
#define Error errhdl_error
#define Escalate() errhdl_error(errhdl_class, "%s", errhdl_message)
#define ErrorClass errhdl_class
#define ErrorOfClass(class) errhdl_errorofclass(errhdl_class, class)
#define ErrorMessage errhdl_message

/*
 *  Several standard error macros are following:
 */
#define SystemError() \
	Error("system", "System error #%i: %s", errno, strerror(errno))
#define MemoryError() \
	Error("memory", "Memory allocation failed.")
#define OverflowError() \
	Error("overflow", "Overflow occured.")
#define ArgumentError() \
	Error("argument", "Illegal argument passed to function.")
#define InternalError() \
	Error("internal", "Internal error")
#define ImplementError() \
	Error("implement", "Function not implemented.")
#define IOError() \
	Error("io", "Input/Output error")
#define DataError() \
	Error("data", "Data could not be processed.")
#define UnknownError() \
	Error("unknown", "An unknown error has occured.")

/*
 *  Macro for functions returning an errno constant (i.e. pthread_create):
 */
#define CheckRetval(x) if ((errno = (x))) SystemError();


#endif


